"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const _ = require("lodash");
const metadata_1 = require("../metadata");
const decorators_1 = require("../decorators");
class DbImpl {
    constructor(modelFn, model) {
        this.modelFn = modelFn;
        this.model = model;
        this.createCopyValsFn();
        this.createTransformQueryFn();
    }
    create(t) {
        return __awaiter(this, void 0, void 0, function* () {
            if (t instanceof Array) {
                t = _.cloneDeep(t);
                t.forEach((t, idx, arr) => { arr[idx] = this.transformQuery(t); });
                let results = yield this.model.bulkCreate(t);
                // return this.wrapResults(results);
                return true;
            }
            else {
                t = this.transformQuery(t);
                let result = yield this.model.create(t);
                return this.wrapResult(result);
            }
        });
    }
    findById(id) {
        return __awaiter(this, void 0, void 0, function* () {
            let result = yield this.model.findById(id);
            return result && this.wrapResult(result);
        });
    }
    findOne(query) {
        return __awaiter(this, void 0, void 0, function* () {
            query = _.clone(query);
            query.where = this.transformQuery(query.where);
            let result = yield this.model.findOne(query);
            return result && this.wrapResult(result);
        });
    }
    findOrCreate(query, defaults) {
        return __awaiter(this, void 0, void 0, function* () {
            query = this.transformQuery(query);
            defaults = this.transformQuery(defaults);
            let [result, created] = yield this.model.findOrCreate({ where: query, defaults: defaults || {} });
            return [result && this.wrapResult(result), created];
        });
    }
    findAndCountAll(query) {
        return __awaiter(this, void 0, void 0, function* () {
            if (query) {
                query = _.clone(query);
                query.where = this.transformQuery(query.where);
            }
            let results = yield this.model.findAndCountAll(query);
            return { count: results.count, results: this.wrapResults(results.rows) };
        });
    }
    findAll(query) {
        return __awaiter(this, void 0, void 0, function* () {
            if (query) {
                query = _.clone(query);
                query.where = this.transformQuery(query.where);
            }
            let results = yield this.model.findAll(query);
            return this.wrapResults(results);
        });
    }
    all(query) {
        return this.findAll(query);
    }
    count(query) {
        return __awaiter(this, void 0, void 0, function* () {
            if (query) {
                query = _.clone(query);
                query.where = this.transformQuery(query.where);
            }
            return yield this.model.count(query);
        });
    }
    max(field) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.model.max(field);
        });
    }
    min(field) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.model.min(field);
        });
    }
    sum(field) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.model.sum(field);
        });
    }
    save(t) {
        return __awaiter(this, void 0, void 0, function* () {
            let [result, created] = yield this.findOrCreate({ id: t.id }, t);
            return result;
        });
    }
    update(query, replace) {
        return __awaiter(this, void 0, void 0, function* () {
            let isId = true;
            if (this.isId(query))
                query = { where: { id: query } };
            else if (this.isT(query))
                query = { where: { id: query.id } };
            else {
                isId = false;
                query.where = this.transformQuery(query.where);
            }
            replace = this.transformQuery(replace);
            let [affected, results] = yield this.model.update(replace, query);
            return affected;
        });
    }
    updateOrCreate(query, defaults) {
        return __awaiter(this, void 0, void 0, function* () {
            let [result, created] = yield this.findOrCreate(query, defaults);
            if (!created) {
                let worked = yield this.update({ where: query }, defaults);
                if (!worked)
                    throw new Error("Failed to update or create a model.");
                let resultOrNull = yield this.findOne(query);
                if (!resultOrNull)
                    throw new Error("Updated row, but could not find it afterwards.");
                result = resultOrNull;
            }
            return [result, created];
        });
    }
    destroy(query) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isId(query))
                query = { where: { id: query } };
            else if (this.isT(query))
                query = { where: { id: query.id } };
            else {
                query = _.cloneDeep(query);
                query.where = this.transformQuery(query.where);
            }
            return yield this.model.destroy(query);
        });
    }
    isId(query) {
        return typeof query === 'number' || typeof query == 'string';
    }
    isT(query) {
        return !!query.id;
    }
    createCopyValsFn() {
        let allProps = [];
        let directTransformFn = (val) => val;
        var props = Reflect.getOwnMetadata(metadata_1.ModelPropertiesSym, this.modelFn.prototype) || [];
        for (var q = 0; q < props.length; q++) {
            var propName = props[q];
            var propMeta = Reflect.getOwnMetadata(metadata_1.PropMetadataSym, this.modelFn.prototype, propName);
            if (!propMeta)
                throw new Error(`Could not find model property metadata for property ${this.modelFn.name || this.modelFn}.${propName}.`);
            let transformFn = directTransformFn;
            if (propMeta.type == decorators_1.Types.DATE)
                transformFn = (dateStr) => new Date(dateStr);
            allProps.push({ columnName: propMeta.columnName || propName, propertyName: propName, transformFn: transformFn });
        }
        // console.log(this.modelFn.name || this.modelFn, 'allProps:', `[\r\n    ${allProps.map(p => p.propertyName + ': ' + p.columnName).join(',\r\n    ')}\r\n]`);
        this.copyVals = function (sql, t) {
            for (var q = 0; q < allProps.length; q++) {
                // console.log(allProps[q].propertyName + ':', sql[allProps[q].propertyName]);
                // t[allProps[q].propertyName] = allProps[q].transformFn(sql[allProps[q].columnName]);
                t[allProps[q].propertyName] = allProps[q].transformFn(sql[allProps[q].propertyName]);
            }
            // console.log(JSON.stringify(sql));
        };
        //TODO: deep copy?
    }
    createTransformQueryFn() {
        let allTransforms = [];
        var belongsTo = Reflect.getOwnMetadata(metadata_1.ModelBelongsToAssociationsSym, this.modelFn.prototype) || [];
        for (var q = 0; q < belongsTo.length; q++) {
            var propName = belongsTo[q];
            var propMeta = Reflect.getOwnMetadata(metadata_1.BelongsToMetadataSym, this.modelFn.prototype, propName);
            if (!propMeta)
                throw new Error(`Could not find model belongs-to metadata for property ${this.modelFn.name || this.modelFn}.${propName}.`);
            let fkey = propMeta.foreignKey;
            if (fkey && typeof fkey !== 'string')
                fkey = fkey.name;
            if (!fkey)
                throw new Error(`Could not get foreign key for property ${this.modelFn.name || this.modelFn}.${propName}`);
            allTransforms.push({
                type: 'belongs-to',
                columnName: fkey,
                fieldName: propName,
                pkName: 'id' // propMeta.targetKey || 'id'
            });
        }
        this.transformQuery = function (query) {
            query = _.clone(query);
            for (var q = 0; q < allTransforms.length; q++) {
                let transform = allTransforms[q];
                switch (transform.type) {
                    case 'belongs-to':
                        if (query[transform.fieldName]) {
                            if (query[transform.fieldName][transform.pkName])
                                query[transform.fieldName] = query[transform.fieldName][transform.pkName];
                            query[transform.columnName] = query[transform.fieldName];
                            delete query[transform.fieldName];
                        }
                        break;
                    default:
                        throw new Error(`WTF? How did you get here?`);
                }
            }
            return query;
        };
        let oldCopyVals = this.copyVals;
        this.copyVals = function (sql, t) {
            oldCopyVals(sql, t);
            for (var q = 0; q < allTransforms.length; q++) {
                let transform = allTransforms[q];
                switch (transform.type) {
                    case 'belongs-to':
                        t[transform.fieldName] = sql[transform.columnName];
                        delete t[transform.columnName];
                        break;
                    default:
                        throw new Error(`WTF? How did you get here?`);
                }
            }
        };
    }
    fromJson(json) {
        return this.wrapResult(json);
    }
    wrapResult(result) {
        let t = new this.modelFn();
        this.copyVals(result, t);
        return t;
    }
    wrapResults(results) {
        return results.map(result => this.wrapResult(result));
    }
}
exports.DbImpl = DbImpl;
//# sourceMappingURL=db-impl.js.map